@@ -1,6 +1,7 @@
+from functools import wraps
 import httpx
 from pydantic import parse_obj_as
-from .types import Beatmap, Beatmaps, Rankings, User, Scores, Score, GameMode, GameModeInt, RankingType, ScoreTypes, BeatmapUserScore, BeatmapScores, Attributes, KudosuHistory
+from .types import Beatmap, Beatmaps, Rankings, User, Scores, Score, GameMode, GameModeInt, RankingType, ScoreTypes, BeatmapUserScore, BeatmapScores, Attributes, KudosuHistory, Event, Users, Beatmapset, BeatmapPlaycount, BeatmapType, Spotlights
 from .utility import c_TypeError
 
 class AsyncOsuApi:
@@ -16,11 +17,10 @@ class AsyncOsuApi:
         self.Client = httpx.AsyncClient()
         self.client_id = client_id
         self.client_secret = client_secret
-        self.token_type, self.access_token = self.__get_auth()
-        self.authorization = self.token_type+" "+self.access_token
+        self.authorization = self.__new_auth()
 
     #? https://osu.ppy.sh/docs/index.html#client-credentials-grant
-    def __get_auth(self):
+    def __new_auth(self):
         body_params = {
             "client_id" : self.client_id,
             "client_secret" : self.client_secret,
@@ -31,21 +31,46 @@ class AsyncOsuApi:
         response = httpx.post(url=self.TOKEN_URL, json=body_params, headers=self.base_headers).json()
         if "error" in response:
             raise ConnectionError(f'error: {response["error"]}')
-        return response["token_type"], response["access_token"]
+        return response["token_type"] +" "+ response["access_token"]
+
+    def __del_auth(self):
+        if self.authorization == None:
+            raise ValueError("self.authorization is type<None> or is not set.")
 
-    #? https://osu.ppy.sh/docs/index.html#lookup-beatmap
-    async def lookup_beatmap(self, 
-                       beatmap_id:int, 
-                       checksum:str=None, 
-                       filename:str=None) -> Beatmap:
         headers = self.base_headers
         headers["Authorization"] = self.authorization
+
+        httpx.delete(url=self.BASE_URL+"/oauth/tokens/current", headers=headers)
+        self.authorization = None
+
+    def __del__(self):
+        self.__del_auth()
+    
+    def __exit__(self):
+        self.__del_auth()
+
+    def verify_auth(func):
+        @wraps(func)
+        def wrapper(self, *args, **kwargs):
+            if not self.authorization:
+                self.authorization = self.__new_auth()
+            headers = self.base_headers
+            headers["Authorization"] = self.authorization
+            return func(self, headers=headers, *args, **kwargs)
+        return wrapper
+    
+    #? https://osu.ppy.sh/docs/index.html#lookup-beatmap
+    @verify_auth
+    async def lookup_beatmap(self,
+                             beatmap_id:int, 
+                             checksum:str=None, 
+                             filename:str=None,
+                             **kwargs) -> Beatmap:
         query_params = {}
 
         if not isinstance(beatmap_id, int):
             raise c_TypeError(param_name="beatmap_id",correct="int",wrong=type(beatmap_id).__name__)
-        else:
-            query_params["id"] = beatmap_id
+        query_params["id"] = beatmap_id
 
         if checksum:
             if not isinstance(checksum, str):
@@ -58,17 +83,17 @@ class AsyncOsuApi:
             query_params["filename"] = filename
 
 
-        response = await self.Client.get(url=self.BASE_URL+"/beatmaps/lookup", params=query_params, headers=headers)
+        response = await self.Client.get(url=self.BASE_URL+"/beatmaps/lookup", params=query_params, headers=kwargs["headers"])
         return parse_obj_as(type_=Beatmap, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-a-user-beatmap-score
-    async def user_beatmap_score(self, 
-                           beatmap_id:int, 
-                           user_id:int, 
-                           mode:GameMode|str=None, 
-                           mods:str=None)->BeatmapUserScore:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def user_beatmap_score(self,
+                                 beatmap_id:int,
+                                 user_id:int,
+                                 mode:GameMode|str=None,
+                                 mods:str=None,
+                                 **kwargs)->BeatmapUserScore:
         query_params = {}
         
         if not isinstance(beatmap_id, int):
@@ -88,16 +113,16 @@ class AsyncOsuApi:
                 raise c_TypeError(param_name="mods",correct="str",wrong=type(mods).__name__)
             query_params["mods"] = mods
 
-        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=BeatmapUserScore, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-a-user-beatmap-scores
-    async def user_beatmap_scores(self, 
-                            beatmap_id:int, 
-                            user_id:int, 
-                            mode:GameMode|str=None)->Scores:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def user_beatmap_scores(self,
+                                  beatmap_id:int, 
+                                  user_id:int, 
+                                  mode:GameMode|str=None,
+                                  **kwargs)->Scores:
         query_params = {}
 
         if not isinstance(beatmap_id, int):
@@ -112,17 +137,17 @@ class AsyncOsuApi:
                 mode = mode.value
             query_params["mode"] = mode
         
-        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}/all", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores/users/{user_id}/all", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=Scores, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-beatmap-scores
-    async def beatmap_scores(self, 
-                       beatmap_id:int, 
-                       mode:GameMode|str=None, 
-                       mods:str=None, 
-                       type:str=None)->BeatmapScores:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def beatmap_scores(self,
+                             beatmap_id:int, 
+                             mode:GameMode|str=None, 
+                             mods:str=None,
+                             Type:str=None,
+                             **kwargs)->BeatmapScores:
         query_params = {}
 
         if not isinstance(beatmap_id, int):
@@ -140,18 +165,18 @@ class AsyncOsuApi:
                 raise c_TypeError(param_name="mods",correct="str",wrong=type(mods).__name__)
             query_params["mods"] = mods
 
-        if type:
+        if Type:
             if not isinstance(type, str):
                 raise c_TypeError(param_name="type",correct="str",wrong=type(type).__name__)
             
-        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/scores", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=BeatmapScores, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-beatmaps
-    async def beatmaps(self, 
-                 beatmap_ids:list[int])->Beatmaps:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def beatmaps(self,
+                       beatmap_ids:list[int],
+                       **kwargs)->Beatmaps:
         query_params = {}
 
         if not isinstance(beatmap_ids, list):
@@ -160,29 +185,29 @@ class AsyncOsuApi:
             raise c_TypeError(param_name="beatmap_ids[elements]", correct="int", wrong=type(beatmap_ids[0]).__name__)
         query_params["ids[]"] = beatmap_ids
 
-        response = await self.Client.get(url=self.BASE_URL+"/beatmaps", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+"/beatmaps", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=Beatmaps, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-beatmap
-    async def beatmap(self, 
-                beatmap_id:int)->Beatmap:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def beatmap(self,
+                      beatmap_id:int,
+                      **kwargs)->Beatmap:
 
         if not isinstance(beatmap_id, int):
             raise c_TypeError(param_name="beatmap_id", correct="int", wrong=type(beatmap_id).__name__)
         
-        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}", headers=headers)
+        response = await self.Client.get(url=self.BASE_URL+f"/beatmaps/{beatmap_id}", headers=kwargs["headers"])
         return parse_obj_as(type_=Beatmap, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-beatmap-attributes
-    async def beatmap_attributes(self, 
-                           beatmap_id:int,
-                           mods:list[str]=None,
-                           ruleset: GameMode|str=None,
-                           ruleset_id: GameModeInt|int=None)->Attributes:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def beatmap_attributes(self,
+                                 beatmap_id:int,
+                                 mods:list[str]=None,
+                                 ruleset: GameMode|str=None,
+                                 ruleset_id: GameModeInt|int=None,
+                                 **kwargs)->Attributes:
         query_params = {}
 
         if not isinstance(beatmap_id, int):
@@ -207,16 +232,16 @@ class AsyncOsuApi:
                 raise c_TypeError(param_name="ruleset_id", correct="GameMode|int",wrong=type(ruleset_id).__name__)
             query_params["ruleset_id"] = ruleset_id
 
-        response = await self.Client.post(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/attributes", headers=headers, params=query_params)
+        response = await self.Client.post(url=self.BASE_URL+f"/beatmaps/{beatmap_id}/attributes", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=Attributes, obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-user-kudosu
-    async def user_kudosu(self, 
-                    user_id:int, 
-                    limit:int=None, 
-                    offset:str=None)->list[KudosuHistory]:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def user_kudosu(self,
+                          user_id:int,
+                          limit:int=None,
+                          offset:str=None,
+                          **kwargs)->list[KudosuHistory]:
         query_params = {}
 
         if not isinstance(user_id, int):
@@ -232,19 +257,19 @@ class AsyncOsuApi:
                 raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
             query_params["offset"] = offset
             
-        response = await self.Client.get(url=self.BASE_URL+f"/users/{user_id}/kudosu", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/users/{user_id}/kudosu", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=list[KudosuHistory], obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-user-scores
-    async def user_scores(self, 
-                    user_id:int, 
-                    type:str, 
-                    include_fails:int=0,
-                    mode:GameMode|str=None,
-                    limit:int=None,
-                    offset:int=None)->list[Score]:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+    @verify_auth
+    async def user_scores(self,
+                          user_id:int,
+                          Type:str,
+                          include_fails:int=0,
+                          mode:GameMode|str=None,
+                          limit:int=None,
+                          offset:int=None,
+                          **kwargs)->list[Score]:
         query_params = {}
         
         if include_fails != 0:
@@ -274,21 +299,70 @@ class AsyncOsuApi:
         response = await self.Client.get(url=self.BASE_URL+f"/users/{user_id}/scores/{type}", headers=self.base_headers, params=query_params)
         return parse_obj_as(type_=list[Score], obj=response.json())
                 
-    # TODO make user_beatmaps request https://osu.ppy.sh/docs/index.html#get-user-beatmaps
-    async def user_beatmaps(self):
-        raise Exception("user_beatmaps not implemented.")
+    #? https://osu.ppy.sh/docs/index.html#get-user-beatmaps
+    @verify_auth
+    async def user_beatmaps(self,
+                            user_id:int,
+                            Type:BeatmapType|str,
+                            limit:int=None,
+                            offset:str=None,
+                            **kwargs)->list[BeatmapPlaycount] | list[Beatmapset]:
+        query_params = {}
+        
+        if not isinstance(user_id, int):
+            raise c_TypeError(param_name="user_id", correct="int", wrong=type(user_id).__name__)
+        if not isinstance(Type, (BeatmapType,str)):
+            raise c_TypeError(param_name="Type", correct="BeatmapType|str", wrong=type(Type).__name__)  
+        if isinstance(Type, BeatmapType):
+            Type = Type.value
+         
+        if limit:
+            if not isinstance(limit, int):
+                raise c_TypeError(param_name="limit", correct="int", wrong=type(limit).__name__)
+            query_params["limit"] = limit
+        
+        if offset:
+            if not isinstance(offset, str):
+                raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
+            query_params["offset"] = offset
+            
+        response = await self.Client.get(url=self.BASE_URL+f"/users/{user_id}/beatmapsets/{Type}", headers=kwargs["headers"], params=query_params)
+        if Type == BeatmapType.MOST_PLAYED.value:
+            return parse_obj_as(type_=list[BeatmapPlaycount], obj=response.json())
+        return parse_obj_as(type_=list[Beatmapset], obj=response.json())
     
-    # TODO make user_recent_activity request https://osu.ppy.sh/docs/index.html#get-user-recent-activity
-    async def user_recent_activity(self):
-        raise Exception("user_recent_activity not implemented.")
+    #? https://osu.ppy.sh/docs/index.html#get-user-recent-activity
+    @verify_auth
+    async def user_recent_activity(self,
+                                   user_id:int,
+                                   limit:int=None,
+                                   offset:str=None,
+                                   **kwargs)->list[Event]:
+        query_params = {}
+        
+        if not isinstance(user_id, int):
+            raise c_TypeError(param_name="user_id", correct="int", wrong=type(user_id).__name__)
+         
+        if limit:
+            if not isinstance(limit, int):
+                raise c_TypeError(param_name="limit", correct="int", wrong=type(limit).__name__)
+            query_params["limit"] = limit
+        
+        if offset:
+            if not isinstance(offset, str):
+                raise c_TypeError(param_name="offset", correct="str", wrong=type(offset).__name__)
+            query_params["offset"] = offset
+
+        response = await self.Client.get(url=self.BASE_URL+f"/users/{user_id}/recent_activity", headers=kwargs["headers"], params=query_params)
+        return parse_obj_as(type_=list[Event], obj=response.json())
     
     #? https://osu.ppy.sh/docs/index.html#get-user
+    @verify_auth
     async def user(self,
-             username:int|str,
-             mode:GameMode|str="",
-             key:str="")-> User:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+                   username:int|str,
+                   mode:GameMode|str="",
+                   key:str="",
+                   **kwargs)-> User:
         query_params = {}
 
         if not isinstance(username, (int,str)):
@@ -305,31 +379,47 @@ class AsyncOsuApi:
                 raise ValueError("key can only be 'id' or 'username'")
             query_params["key"] = key
 
-        response = await self.Client.get(url=self.BASE_URL+f"/users/{username}/{mode}", headers=headers,params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/users/{username}/{mode}", headers=kwargs["headers"],params=query_params)
         return parse_obj_as(type_=User, obj=response.json())
 
-    # TODO make users request https://osu.ppy.sh/docs/index.html#get-users
-    async def users(self):
-        raise Exception("users not implemented")
+    #? https://osu.ppy.sh/docs/index.html#get-users
+    @verify_auth
+    async def users(self,
+                    user_ids:list[int],
+                    **kwargs)->Users:
+        query_params = {}
+
+        if not isinstance(user_ids, list):
+            raise c_TypeError(param_name="user_ids", correct="list", wrong=type(user_ids).__name__)
+        if not isinstance(user_ids[0], int):
+            raise c_TypeError(param_name="user_ids[elements]", correct="int", wrong=type(user_ids[0]).__name__)
+        query_params["ids[]"] = user_ids
+
+        response = await self.Client.get(url=self.BASE_URL+"/users", headers=kwargs["headers"], params=query_params)
+        return parse_obj_as(type_=Users, obj=response.json())
 
     #? https://osu.ppy.sh/docs/index.html#get-ranking
+    @verify_auth
     async def ranking(self,
-                mode:GameMode|str,
-                type:str,
-                filter:str="all",
-                country:int=None,
-                cursor:int=None,
-                spotlight_id:int=None,
-                variant:str=None)->Rankings:
-        headers = self.base_headers
-        headers["Authorization"] = self.authorization
+                      mode:GameMode|str,
+                      Type:str,
+                      filter:str="all",
+                      country:int=None,
+                      cursor:int=None,
+                      spotlight_id:int=None,
+                      variant:str=None,
+                      **kwargs)->Rankings:
         query_params = {}
 
         if not isinstance(mode, (GameMode, str)):
             raise c_TypeError(param_name="mode",correct="GameMode|str",wrong=type(mode).__name__)
         if isinstance(mode, GameMode):
             mode = mode.value
-        query_params["mode"] = mode
+
+        if not isinstance(Type, (RankingType, str)):
+            raise c_TypeError(param_name="Type", correct="RankingType|str", wrong=type(Type).__name__)
+        if isinstance(Type, RankingType):
+            Type = Type.value
 
         if filter not in ["all","friends"]:
             raise ValueError("filter must be 'all' or 'friends'")
@@ -363,9 +453,12 @@ class AsyncOsuApi:
                 raise ValueError("mode must be 'mania' to use variant")
             query_params["variant"] = variant
 
-        response = await self.Client.get(url=self.BASE_URL+f"/rankings/{mode}/{type}", headers=headers, params=query_params)
+        response = await self.Client.get(url=self.BASE_URL+f"/rankings/{mode}/{type}", headers=kwargs["headers"], params=query_params)
         return parse_obj_as(type_=Rankings, obj=response.json())
     
-    # TODO make spotlights request https://osu.ppy.sh/docs/index.html#get-spotlights
-    async def spotlights(self):
-        raise Exception("spotlights not implemented.")
\ No newline at end of file
+    #? https://osu.ppy.sh/docs/index.html#get-spotlights
+    @verify_auth
+    async def spotlights(self, **kwargs)->Spotlights:
+
+        response = await self.Client.get(url=self.BASE_URL+"/spotlights", headers=kwargs["headers"])
+        return parse_obj_as(type_=Spotlights, obj=response.json())
\ No newline at end of file
