@@ -1,6 +1,8 @@
+from datetime import datetime
+from typing import Any
 from pydantic import BaseModel, Field
-from typing import Optional
 from .Extras import Availability, Nominations, Hype
+from .User import UserCompact
 from .Covers import Covers
 
 # https://osu.ppy.sh/docs/index.html#beatmapsetcompact
@@ -20,8 +22,23 @@ class BeatmapsetCompact(BaseModel):
     title_unicode: str
     user_id: int
     video: bool
-    ratings: list[int]
-
+    
+    # optionals
+    beatmaps: list[Any] | None
+    converts: Any | None
+    current_user_attributes: Any | None
+    description: Any | None
+    discussions: Any | None
+    events: Any | None
+    genre: Any | None
+    has_favourited: bool | None
+    language: Any | None
+    nominations: Any | None
+    ratings: list[int] | None
+    recent_favourites: Any | None
+    related_users: Any | None
+    user: UserCompact | None
+    
     class Config:
         allow_population_by_field_name = True
         arbitrary_types_allowed = True
@@ -33,18 +50,21 @@ class Beatmapset(BeatmapsetCompact):
     can_be_hyped: bool
     creator: str
     discussion_locked: bool
-    hype: Optional[Hype] = None
+    hype: Hype | None
     is_scoreable: bool
-    last_updated: str
-    legacy_thread_url: Optional[str] = None
-    nominations: Nominations = Field(alias="nominations_summary")
+    last_updated: datetime
+    legacy_thread_url: str | None
+    nominations: Nominations | None
     ranked: int
-    ranked_date: Optional[str] = None
+    ranked_date: datetime | None
     source: str
     storyboard: bool
-    submitted_date: Optional[str] = None
+    submitted_date: datetime | None
     tags: str
 
     class Config:
         allow_population_by_field_name = True
-        arbitrary_types_allowed = True
\ No newline at end of file
+        arbitrary_types_allowed = True
+        json_encoders = {
+            datetime: str,
+		}
\ No newline at end of file
