@@ -1,12 +1,13 @@
-from .Beatmap import Beatmap, BeatmapCompact, Beatmaps
+from .Beatmap import Beatmap, BeatmapCompact, Beatmaps, BeatmapPlaycount
 from .Beatmapset import Beatmapset, BeatmapsetCompact
 from .Covers import Covers
 from .DifficultyAttributes import DifficultyAttributes
-from .Enums import UserAcountHistoryTypes, UserListFilters, UserListSorts, UserListViews, GameMode, GameModeInt, ScoreTypes, RankingType, BeatmapsetDownload
+from .Enums import UserAcountHistoryTypes, UserListFilters, UserListSorts, UserListViews, GameMode, GameModeInt, ScoreTypes, RankingType, BeatmapsetDownload, BeatmapType
 from .Extras import GradeCounts, Level, Cover, Country, RankHighest, RankHistory, ReplaysWatchedCount, ProfileBanner, Availability, Hype, Nominations, Weight, Page, Kudosu, Statistics, ScoreMatchInfo, Attributes, Giver, Post, Failtimes
 from .KudosuHistory import KudosuHistory
 from .Rankings import Spotlight, Spotlights, Cursor, Rankings
 from .Score import Score, BeatmapScores, BeatmapUserScore, Scores
-from .User import User, UserCompact
+from .User import User, UserCompact, Users
 from .UserExtras import UserGroup, UserAccountHistory, UserBadge, UserMonthlyPlaycount, UserAchievement, UserProfileCustomization
-from .UserStatistics import UserStatistics, UserStatisticsRulesets
\ No newline at end of file
+from .UserStatistics import UserStatistics, UserStatisticsRulesets
+from .Event import Event
\ No newline at end of file
